import pytest

from ejudge import functions
from ejudge.tests import abstract as base

sources = r"""
## ok
name = raw_input("name: ")
print("hello %s!" % name)

## wrong
name = raw_input("name: ")
print(name)

## syntax
a b

## error
1 / 0

## recursive
def f(x):
    return 1 if x <= 1 else x * f(x - 1)
print(f(5))
"""


@pytest.mark.python2
class TestPython2Support(base.TestLanguageSupport):
    base_lang = 'python2'
    source_all = sources

    @pytest.mark.skip
    def test_error_feedback_message_shows_traceback(self, lang, iospec):
        iospec = functions.run('1 / 0', ['foo'], lang=lang, sandbox=False)
        assert iospec[0].error_message == (
            'Traceback (most recent call last)\n'
            '  File "main.py", line 1, in <module>\n'
            'ZeroDivisionError: division by zero'
        )



import pytest

import ejudge
import iospec
from ejudge import functions

simple_inputs = """
<foo>
<bar>
<baz>
foo bar baz
"""

simple_source = """
x, y, z = input(), input(), input()
print(x, y, z)
"""


# Program uses less inputs than predicted on iospec file
def test_less_inputs_than_expected(lang='python', sandbox=False):
    src = 'print(42)'
    iosrc = (
        'foo<42>\n'
        'bar<0>\n'
        '42'
    )
    fb = functions.grade(src, iosrc, lang=lang, sandbox=sandbox)
    fb.testcase.pprint()
    fb.answer_key.pprint()
    assert fb.grade == 0
    assert fb.status == 'wrong-answer'


@pytest.mark.sandbox
def test_less_inputs_than_expected_sandbox():
    test_less_inputs_than_expected(sandbox=True)


def test_less_inputs_than_expected_pytuga():
    test_less_inputs_than_expected(lang='pytuga')


# Simple inputs with out the corresponding output interaction
def test_simple_inputs():
    ast = iospec.parse(simple_inputs)
    ast.normalize()
    result = functions.run(simple_source, ast, lang='python', sandbox=False)
    result.normalize()
    ast.pprint()
    result.pprint()
    assert ast.to_json() == result.to_json()
    assert iospec.isequal(ast, result)


def test_grade_simple_inputs():
    ast = iospec.parse(simple_inputs)
    feedback = functions.grade(simple_source, ast, lang='python', sandbox=False)
    assert feedback.grade == 1


def test_grade_program_with_simple_inputs():
    io = """<1>
<2>
<3>
6
"""
    src = """
print(sum([int(input()) for _ in range(3)]))
"""
    feedback = ejudge.grade(src, io, lang='python', sandbox=False)
    assert feedback.grade == 1


import builtins
import pytuga
from ejudge.build_manager import IntegratedBuildManager
from ejudge.execution_manager import IntegratedExecutionManager


class PytugaBuildManager(IntegratedBuildManager):
    """
    Build manager for the Pytuga language.
    """

    language = 'pytuga'
    transpiled = ''


    def syntax_check(self):
        pytuga.compile(self.source, 'main.pytg', 'exec')

    def build(self):
        super().build()
        self.is_built = False
        self.transpiled = pytuga.transpile(self.source)
        self.is_built = True

    def get_modules(self):
        return super().get_modules() + ['pytuga.lib.forbidden']


class PytugaExecutionManager(IntegratedExecutionManager):
    """
    Execution manager for the Pytugues language.
    """

    language = 'pytuga'

    @property
    def transpiled(self):
        return self.build_manager.transpiled

    def exec(self, globals, locals):
        if globals is None:
            globals = {}
        globals.update(pytuga.tugalib_namespace(forbidden=True))
        builtins.exec(self.transpiled, globals, locals)


import subprocess
import tempfile

import shutil

from ejudge.build_manager import CompiledLanguageBuildManager
from ejudge.execution_manager import CompiledLanguageExecutionManager


#
# C compilers
#
# We have manager classes for the two biggest compilers gcc and clang and also
# for tcc, the tiny C compiler, which is a small and fast compiler. In the
# future we might create an IntegratedManager based on clang or tcc or even on
# a C interpreter like ROOT's Cint.
#
class CLanguageBuildManager(CompiledLanguageBuildManager):
    """
    Common functionality for all C-based builders.
    """

    source_extension = '.c'
    language = 'c'
    shell_checker_args = ['gcc', '-fsyntax-only']

    def syntax_check(self):
        c_syntax_check(self.source, compiler='gcc')


class CLanguageExecutionManager(CompiledLanguageExecutionManager):
    """
    Execution manager for all C-based languages.
    """


class GccBuildManager(CLanguageBuildManager):
    """
    Build executable with GCC.
    """

    language = 'c'
    build_args = ['gcc', '-lm', 'main.c', '-std=c99', '-o', 'main.exe']


class TccBuildManager(CLanguageBuildManager):
    """
    Build executable with TCC.
    """

    language = 'tcc'
    build_args = ['tcc', '-w', '-o', 'main.exe', 'main.c']

    def syntax_check(self):
        c_syntax_check(self.source, compiler='tcc')


class ClangBuildManager(CLanguageBuildManager):
    """
    Build executable with Clang compiler.
    """

    language = 'clang'
    build_args = ['clang', '-lm', 'main.c', '-std=c99', '-o', 'main.exe']

    def syntax_check(self):
        c_syntax_check(self.source, compiler='clang')


#
# C++ compilers
#
class CppBuildManager(CompiledLanguageBuildManager):
    """
    Base class for C++ builders.
    """

    source_extension = '.cpp'
    language = 'c++'


class GccCppBuildManager(CppBuildManager):
    """
    Build executable with g++
    """

    language = 'g++'
    build_args = ['g++', '-lm', '-o', 'main.cpp', '-o', 'main.exe']

    def syntax_check(self):
        c_syntax_check(self.source, compiler='g++', cpp=True)


class ClangCppBuildManager(CppBuildManager):
    """
    Build executable from C++ source with clang.
    """

    language = 'clang++'
    build_args = ['clang++', '-lm', '-o', 'main.cpp', '-o', 'main.exe']

    def syntax_check(self):
        c_syntax_check(self.source, compiler='clang++', cpp=True)


def c_syntax_check(source, compiler=None, cpp=False, encoding='utf8'):
    """
    Check syntax of C code.

    Raises a SyntaxError on error. Do not return anything.

    Args:
        source (str):
            A string with C or C++ source code.
        compiler (optional, str or list):
            One of {clang, gcc, tcc} (or their corresponding C++ counterparts)
            or a list of priorities.
        cpp (bool):
            Set to true for C++ syntax check.
        encoding (str):
            Encoding for the program source.
    """

    compilers_c = ['clang', 'gcc', 'tcc']
    compilers_cpp = ['clang++', 'g++']
    compilers = compilers_cpp if cpp else compilers_c

    # Create list of compilers by priority
    if isinstance(compiler, str):
        if compiler not in compilers:
            raise ValueError('invalid compiler: %r' % compiler)

        compilers.pop(compilers.index(compiler))
        compilers = [compiler] + compilers
    elif compiler is not None:
        compiler = list(compiler)

        for comp_item in compiler:
            if comp_item not in compilers:
                raise ValueError('invalid compiler: %r' % compiler)

        compiler.extend(x for x in compilers if x not in compiler)
        compilers = compiler

    # Find the first compiler executable installed in the system
    for compiler in compilers:
        compiler_path = shutil.which(compiler)
        if compiler_path:
            break
    else:
        msg = 'no suitable compiler found from list %r' % compilers
        raise RuntimeError(msg)

    # Open a temporary file and execute ``CC -fsytnax-only source.c``
    with tempfile.NamedTemporaryFile(mode='w',
                                     suffix='.cpp' if cpp else '.c',
                                     encoding=encoding,
                                     delete=False) as F:
        F.write(source)
        F.flush()
        cmd = [compiler_path, '-fsyntax-only', F.name, '-o', '/dev/null']

        try:
            out = subprocess.check_output(cmd, stderr=subprocess.STDOUT)
            out = None
        except subprocess.CalledProcessError as ex:
            out = ex.output.decode('utf8') or 'syntax error'

    if out is not None:
        raise SyntaxError(out)



import pytest
import time

from ejudge import functions
from ejudge.tests import abstract as base

sources = r"""
## ok
name = input("name: ")
print("hello %s!" % name)


## wrong
name = input("name: ")
print(name)


## syntax
a b


## error
1 / 0


## recursive
def f(x):
    return 1 if x <= 1 else x * f(x - 1)
print(f(5))


## timeout
import time

name = input('name: ')
t0 = time.time()

# semi-infinite loop :)
while time.time() < t0 + 0.5:
    pass
"""


@pytest.mark.python
class TestPythonSupport(base.TestLanguageSupport):
    base_lang = 'python'
    source_all = sources

    def test_error_feedback_message_shows_traceback(self, lang, iospec):
        feedback = functions.grade('1 / 0', iospec, lang=lang, sandbox=False)
        assert feedback.grade == 0
        assert feedback.title == 'Runtime Error'
        assert feedback.testcase.error_message == (
            'Traceback (most recent call last)\n'
            '  File "main.py", line 1, in <module>\n\n'
            'ZeroDivisionError: division by zero')

    def test_raises_timeout_error(self, lang, iospec):
        src = self.get_source('timeout')
        t0 = time.time()
        result = functions.run(src, iospec, lang=lang, timeout=0.1, sandbox=False)
        dt = time.time() - t0
        assert result.get_error_type() == 'timeout'
        assert dt < 0.45  # slightly less than maximum program runtime

    @pytest.mark.sandbox
    def test_raises_timeout_error_in_sandbox(self, lang, iospec):
        src = self.get_source('timeout')
        result = functions.run(src, iospec, lang=lang, timeout=0.05, sandbox=True)
        assert result.get_error_type() == 'timeout'



import pytest

from ejudge import functions
from ejudge.langs.c_family import c_syntax_check
from ejudge.tests import abstract as base

sources = r"""
## ok
#include<stdio.h>

int main(void) {
    char buffer[100];
    printf("name: ");
    scanf("%s", buffer);
    printf("hello %s!\n", buffer);
}


## wrong
#include<stdio.h>

int main(void) {
    char buffer[100];
    puts("name: ");
    scanf("%s", buffer);
    printf("%s\n", buffer);
}


## syntax
a b

## error
int main(void) {
    char buffer[10];
    char c = buffer[1000];
}


## recursive
#include<stdio.h>

int f(int x) {
    return (x <= 1)? 1: x * f(x - 1);
}

int main(void) {
    printf("%d\n", f(5));
}


## twoinputs
#include<stdio.h>
void main() {
    char name[100], job[200];
    printf("name: ");
    scanf("%s", name);
    printf("job: ");
    scanf("%s", job);
    printf("%s, %s\n", name, job);
}

"""


@pytest.fixture(params=['clang', 'gcc', 'tcc'])
def compiler(request):
    return request.param


@pytest.mark.c
@pytest.mark.gcc
class TestGCCSupport(base.TestLanguageSupport):
    base_lang = 'c'
    source_all = sources

    @pytest.fixture
    def twoinputs(self):
        return self.get_source('twoinputs')

    def test_c_program_with_two_inputs(self, twoinputs):
        result = functions.run(twoinputs, ['foo', 'bar'], lang='c',
                               sandbox=False)
        case = result[0]
        print(result.to_json())
        result.pprint()
        assert list(case) == ['name: ', 'foo', 'job: ', 'bar', 'foo, bar']


@pytest.mark.c
@pytest.mark.gcc
def test_c_syntax_check_detect_invalid_syntax(compiler):
    good_src = TestGCCSupport.get_source('ok')
    bad_src = good_src + '\n}'

    assert c_syntax_check(good_src, compiler=compiler) is None
    with pytest.raises(SyntaxError):
        c_syntax_check(bad_src, compiler=compiler)



import decimal
import io
import traceback

from boxed.jsonbox import run as run_sandbox
from ejudge import registry
from ejudge.exceptions import BuildError
from ejudge.util import keep_cwd, do_nothing_context_manager
from iospec import parse as ioparse, TestCase, ErrorTestCase, IoSpec
from iospec.feedback import feedback as get_feedback


def run(source, inputs, lang=None, *,
        fast=False, timeout=None, raises=False, path=None, sandbox=True,
        **kwargs):
    """
    Run program with the given list of inputs and returns the corresponding
    :class:`iospec.IoSpec` instance with the results.

    Args:
        source (str or file)
            The source code for the test program
        inputs (sequence)
            A sequence of input strings. If input is a sequence of sequences,
            this function will perform multiple test cases. It can also be a
            IoSpec or a TestCase instance which are used to extract the
            necessary input strings.
        lang (str)
            The name for the source code language. See
            :func:`ejudge.graders.io.grade` for more details.
        timeout (float)
            A time limit for the entire run (in seconds). If this attribute is
            not given, the program will run without any timeout. This can be
            potentially dangerous if the input program has an infinite loop.
        sandbox (bool)
            Controls if code is run in sandboxed mode or not. Sandbox protection
            is the default behavior on supported platforms.
        raises (bool)
            If True, raise a BuildError if the build process fails. The default
            behavior is to return a IoSpec instance with a single ErrorTestCase
            with a type string of 'error-build'.
        path (str)
            The absolute file path for the input string or file object.
        fast (bool):
            If true, stops running at the first error.
    Returns:
        A :class:`iospec.IoSpec` structure. If ``inputs`` is a sequence of
        strings, the resulting tree will have a single test case.
    """

    # Normalize inputs
    if isinstance(inputs, (IoSpec, TestCase)):
        inputs = inputs.inputs()
    else:
        if inputs and isinstance(inputs[0], str):
            inputs = [list(inputs)]
        else:
            inputs = [list(map(str, x)) for x in inputs]

    # Optional arguments
    running_in_sandbox = kwargs.get('_running_in_sandbox', False)
    return_json = kwargs.get('_return_json', False)

    # Create build manager
    build_manager = registry.build_manager_from_path(
        lang, source, path,
        is_sandboxed=running_in_sandbox
    )

    # Run in sandboxed mode
    if sandbox:
        imports = build_manager.get_modules()
        result = run_sandbox(
            run,
            args=(source, inputs, lang),
            kwargs={
                'raises': raises,
                'timeout': timeout,
                'fast': fast,
                'path': path,
                'sandbox': False,
                '_running_in_sandbox': True,
                '_return_json': True
            },
            imports=imports,
            timeout=None,
        )
        return IoSpec.from_json(result)

    # Prepare build manager
    try:
        build_manager.build()
    except BuildError as ex:
        if raises:
            raise
        result = IoSpec([ErrorTestCase.build(error_message=str(ex))])
        return result.to_json() if return_json else result

    # Run all examples with the execution manager
    data = []
    language = build_manager.language
    for input_strings in inputs:
        ctrl = registry.execution_manager(language, build_manager,
                                          input_strings)
        try:
            if timeout is None:
                result = ctrl.run()
            elif timeout > 0:
                result = ctrl.run_with_timeout(timeout)
            else:
                raise ValueError('timeout must be positive, got: %s' % timeout)
        except Exception as ex:
            if raises:
                raise
            result = _error_test_case(ex, ex.__traceback__)
        data.append(result)
        if fast and result.is_error:
            break

    # Prepare resulting iospec object
    result = IoSpec(data)
    result.set_meta('lang', build_manager.language)
    if return_json:
        return result.to_json()
    else:
        return result


def grade(source, iospec, lang=None, *,
          fast=True, path=None, raises=False, sandbox=False, timeout=None):
    """
    Grade the string of source code by comparing the results of all inputs and
    outputs in the given template structure.


    Args:
        source (str or file object)
            The source string for the code or a file object
        iospec (IOSpec parse tree)
            The expected template for correct answers.
        lang (str)
            Programming language for the given source code. Users can implement
            plugins to support additional languages or to override the default
            behavior or accepted languages.

            +-----------+------------------------------------------------------+
            | Value     | Description                                          |
            +===========+======================================================+
            | python    | For Python 3.x code. Default runner.                 |
            +-----------+------------------------------------------------------+
            | python2   | Executes in a separate Python 2 interpreter.         |
            +-----------+------------------------------------------------------+
            | tcc       | Compile C code with the tiny C compiler              |
            +-----------+------------------------------------------------------+
            | clang     | Compile C code with clang                            |
            +-----------+------------------------------------------------------+
            | gcc, c    | Compile C code with gcc                              |
            +-----------+------------------------------------------------------+

        sandbox (bool)
            If True, code will run in a sandboxed environment as the `nobody`
            user. It is necessary to have your system properly configured in
            order to do this.
        timeout (float)
            Maximum time (in seconds) for the complete test to run.

    Returns:
        A :class:`ejudge.Feedback` instance.
    """

    if isinstance(iospec, str):
        iospec = ioparse(iospec)
    result = run(source, iospec, lang, fast=fast, path=path, raises=raises,
                 sandbox=sandbox, timeout=timeout)
    feedback = None
    value = decimal.Decimal(1)

    for case, answer_key in zip(result, iospec):
        answer_key = answer_key.copy()
        answer_key.normalize()
        case.normalize()

        curr_feedback = get_feedback(case, answer_key)
        if feedback is None:
            feedback = curr_feedback
        if curr_feedback.grade < value:
            feedback = curr_feedback
            value = curr_feedback.grade
            if value == 0:
                break
    return feedback


def exec(source, lang=None, path=None):
    """
    Execute code in the given language.

    Do not pass any inputs or collect the outputs.
    """

    # Create build manager
    build_manager = registry.build_manager_from_path(
        lang, source, path,
        is_sandboxed=False,
    )
    build_manager.build()
    ctrl = registry.execution_manager(lang, build_manager)
    ctrl.run_interactive()


def _error_test_case(exc, tb, limit=None):
    """
    Return an IoSpec data with a single ErrorTestCase.

    Construct the testcase from the given traceback.
    """

    out = io.StringIO()
    traceback.print_tb(tb, file=out)
    _, sep, tail = out.getvalue().partition('  File "main.py", line')
    message = ('Traceback (most recent call last)\n%s%s: %s' %
               (sep + tail, exc.__class__.__name__, exc))
    return ErrorTestCase.runtime(error_message=message)



import pytest

from ejudge import functions, registry
from iospec import parse as parse_string, datatypes, SimpleTestCase
from iospec.exceptions import BuildError


def source_property(name):
    @property
    def source_property(self):
        return self.get_source(name)
    return source_property


class TestLanguageSupport:
    base_iospec_source = (
        'name: <foo>\n'
        'hello foo!\n'
        '\n'
        'name: <bar>\n'
        'hello bar!'
    )
    base_lang = None
    source_all = None
    source_ok = source_property('ok')
    source_wrong = source_property('wrong')
    source_syntax = source_property('syntax')
    source_recursive = source_property('recursive')
    source_error = source_property('error')

    @pytest.fixture
    def iospec(self):
        return parse_string(self.base_iospec_source)

    @pytest.fixture
    def lang(self):
        return self.base_lang

    @pytest.fixture
    def src_ok(self):
        return self.source_ok

    @pytest.fixture
    def src_wrong(self):
        return self.source_wrong

    @pytest.fixture
    def src_syntax(self):
        return self.source_syntax

    @pytest.fixture
    def src_recursive(self):
        return self.source_recursive

    @pytest.fixture
    def src_error(self):
        return self.source_error

    @pytest.fixture
    def manager_cls(self, lang):
        return registry.build_manager_class(lang)

    @pytest.fixture
    def ex_manager_cls(self, lang):
        return registry.execution_manager_class(lang)

    #
    # Auxiliary functions
    #
    @classmethod
    def get_source(cls, name):
        if cls.source_all is None:
            return None

        _, sep, data = cls.source_all.partition('## %s\n' % name)
        if not sep:
            return None
        data, _, _ = data.partition('\n## ')
        return data.strip()

    #
    # Test simple io.run() interactions
    #
    def test_run_valid_source(self, src_ok, lang, timeout=None, sandbox=False):
        tree = functions.run(src_ok, ['foo'], lang=lang, sandbox=sandbox,
                             timeout=timeout,
                             raises=True)
        try:
            assert len(tree) == 1
            case = tree[0]
            assert isinstance(case, SimpleTestCase)
            assert case[0] == 'name: '
            assert case[1] == 'foo'
            assert case[2] == 'hello foo!'
        except Exception:
            tree.pprint()
            raise

    def test_run_valid_source_with_timeout(self, src_ok, lang):
        self.test_run_valid_source(src_ok, lang, timeout=1.0)

    @pytest.mark.sandbox
    def test_run_valid_source_with_sandbox(self, src_ok, lang):
        self.test_run_valid_source(src_ok, lang, sandbox=True)

    @pytest.mark.sandbox
    def test_run_valid_source_with_sandbox_and_timeout(self, src_ok, lang):
        self.test_run_valid_source(src_ok, lang, sandbox=True, timeout=1.0)

    def test_run_source_with_runtime_error(self, src_error, lang):
        tree = functions.run(src_error, ['foo'], lang=lang, sandbox=False)
        assert tree[0].error_type == 'runtime'

    def test_run_from_input_sequence(self, src_ok, lang):
        inputs = [['foo'], ['bar']]
        tree = functions.run(src_ok, inputs, lang=lang, sandbox=False)
        assert len(tree) == 2
        assert tree[0][2] == 'hello foo!'
        assert tree[1][2] == 'hello bar!'

    def test_run_valid_source_from_iospec_input(self, src_ok, lang):
        case1 = datatypes.SimpleTestCase([datatypes.In('foo'), datatypes.Out('foo')])
        case2 = datatypes.InputTestCase([datatypes.In('bar')])
        inpt = datatypes.IoSpec([case1, case2])
        tree = functions.run(src_ok, inpt, lang=lang, sandbox=False)
        assert len(tree) == 2
        assert tree[0][2] == 'hello foo!'
        assert tree[1][2] == 'hello bar!'

    def test_run_code_with_syntax_error(self, src_syntax, lang):
        ast = functions.run(src_syntax, ['foo'], lang=lang, sandbox=False)
        assert len(ast) == 1
        assert isinstance(ast[0], datatypes.ErrorTestCase)
        assert ast[0].error_type == 'build'

    def test_run_recursive_function(self, src_recursive, lang):
        result = functions.run(src_recursive, [()], lang=lang, sandbox=False)
        assert list(result[0]) == ['120']

    @pytest.mark.sandbox
    def test_run_recursive_function_in_sandbox(self, src_recursive, lang):
        result = functions.run(src_recursive, [()], lang=lang, sandbox=True)
        assert list(result[0]) == ['120']

    #
    # Test grading and check if the feedback is correct
    #
    def test_valid_source_receives_maximum_grade(self, iospec, src_ok, lang):
        feedback = functions.grade(src_ok, iospec, lang=lang, sandbox=False)
        assert isinstance(feedback.answer_key, datatypes.TestCase)
        assert isinstance(feedback.testcase, datatypes.TestCase)
        assert feedback.grade == 1
        assert feedback.message is None
        assert feedback.status == 'ok'

    def test_wrong_source_receives_null_grade(self, iospec, src_wrong, lang):
        feedback = functions.grade(src_wrong, iospec, lang=lang, sandbox=False)
        assert feedback.grade == 0
        assert feedback.status == 'wrong-answer'
        assert feedback.title == 'Wrong Answer'

    #
    # Test build managers
    #
    def test_build_manager_has_the_correct_language(self, manager_cls, src_ok, lang):
        manager = manager_cls(src_ok)
        assert manager.language == lang

    def test_build_manager_can_build(self, manager_cls, src_ok):
        manager = manager_cls(src_ok)
        assert manager.is_built is False
        manager.build()
        assert manager.is_built is True

    def test_build_manager_raise_build_error(self, manager_cls, src_syntax):
        manager = manager_cls(src_syntax)
        with pytest.raises(BuildError):
            manager.build()

    def test_build_manager_can_build_in_sandbox(self, manager_cls, src_ok):
        manager = manager_cls(src_ok, is_sandboxed=True)
        assert manager.is_sandboxed
        assert manager.is_built is False
        manager.build()
        assert manager.is_built is True

    def test_build_manager_raise_build_error(self, manager_cls, src_syntax):
        manager = manager_cls(src_syntax)
        with pytest.raises(BuildError):
            manager.build()

    def test_build_manager_detect_syntax_error_before_build(self, manager_cls, src_syntax, src_ok):
        manager = manager_cls(src_ok)
        manager.syntax_check()  # nothing happens

        manager = manager_cls(src_syntax)
        with pytest.raises(SyntaxError):
            manager.syntax_check()
